/* *************************************************************************
                          boundingbox.cpp  -  description
                             -------------------
    begin                : Wed Oct 16 2002
    copyright            : (C) 2002 by Micha Riser
    email                : mriser@gmx.net
 ************************************************************************* */

/* *************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ************************************************************************* */

#include "boundingbox.h"
#include "intersection.h"

HitStatistics BoundingBox::stat("Bounding Box","Tests","Hits");

void BoundingBox::scale(const Vector3& v) {

    for(int i=0;i<3;i++) {
        
        boxmin.a[i]*=v.a[i];
        boxmax.a[i]*=v.a[i];
            
        if (v.a[i]<0) {                
            // swap min/max
            double tmp = boxmin.a[i];
            boxmin.a[i]=boxmax.a[i];
            boxmax.a[i]=tmp;
        }
        
    }

}


void BoundingBox::updateBoundingBoxPoints(Vector3& min, Vector3& max, const Matrix& rot, Vector3 test) {
    test.applyFromLeft(rot);
    min.minimize(test);
    max.maximize(test);
}


void BoundingBox::rotate(const Matrix& rot) {

    Vector3 t = boxmin;
    t.applyFromLeft(rot);
    Vector3 newmin = t;
    Vector3 newmax = t;

    updateBoundingBoxPoints(newmin, newmax, rot,
      Vector3(boxmin.a[0],boxmin.a[1],boxmax.a[2]) );
    updateBoundingBoxPoints(newmin, newmax, rot,
      Vector3(boxmin.a[0],boxmax.a[1],boxmin.a[2]) );
    updateBoundingBoxPoints(newmin, newmax, rot,
      Vector3(boxmin.a[0],boxmax.a[1],boxmax.a[2]) );
    updateBoundingBoxPoints(newmin, newmax, rot,
      Vector3(boxmax.a[0],boxmin.a[1],boxmin.a[2]) );
    updateBoundingBoxPoints(newmin, newmax, rot,
      Vector3(boxmax.a[0],boxmin.a[1],boxmax.a[2]) );
    updateBoundingBoxPoints(newmin, newmax, rot,
      Vector3(boxmax.a[0],boxmax.a[1],boxmin.a[2]) );
    updateBoundingBoxPoints(newmin, newmax, rot,
      Vector3(boxmax.a[0],boxmax.a[1],boxmax.a[2]) );
        
    boxmin = newmin;
    boxmax = newmax;
    
}


void BoundingBox::transformate(const Transformatable &t) {

    if (t.getTransformationMatrix()) rotate(*t.getInverseMatrix());
    translate(t.getTranslation());

}


DBL BoundingBox::intersect(const Ray &r) const {

    stat.eval();

    // returns INTERSECTION_TIME_EPSILON if ray's origin inside box!

    // calculate x time interval
    DBL tmin;
    DBL tmax;

    if (r.direction.a[0]>0) {

        tmax = (-r.origin.a[0]+boxmax.a[0]) / r.direction.a[0];
        if (tmax<INTERSECTION_TIME_EPSILON) return -1;            
        tmin = (-r.origin.a[0]+boxmin.a[0]) / r.direction.a[0];
        if (tmin<INTERSECTION_TIME_EPSILON) tmin = INTERSECTION_TIME_EPSILON;
            
    } else if (r.direction.a[0]!=0) {

        tmax = (-r.origin.a[0]+boxmin.a[0]) / r.direction.a[0];
        if (tmax<INTERSECTION_TIME_EPSILON) return -1;
        tmin = (-r.origin.a[0]+boxmax.a[0]) / r.direction.a[0];
        if (tmin<INTERSECTION_TIME_EPSILON) tmin = INTERSECTION_TIME_EPSILON;

    } else {

        if (r.origin.a[0]<boxmin.a[0] || r.origin.a[0]>boxmax.a[0] ) return -1;
        tmin = INTERSECTION_TIME_EPSILON;
        tmax = INTERSECTION_INFINIT_TIME;
        
    }
        
    // calculate y time interval
    if (r.direction.a[1]>0) {

        DBL t = (-r.origin.a[1]+boxmax.a[1]) / r.direction.a[1]; //y max
        if (t<tmin) return -1;  if (t<tmax) tmax = t;
        t = (-r.origin.a[1]+boxmin.a[1]) / r.direction.a[1]; //y min
        if (t>tmax) return -1;  if (t>tmin) tmin = t;
            
    } else if (r.direction.a[1]!=0) {

        DBL t = (-r.origin.a[1]+boxmin.a[1]) / r.direction.a[1]; //y max
        if (t<tmin) return -1;  if (t<tmax) tmax = t;
        t = (-r.origin.a[1]+boxmax.a[1]) / r.direction.a[1]; //y min
        if (t>tmax) return -1;  if (t>tmin) tmin = t;
            
    } else {
        
        if (r.origin.a[1]<boxmin.a[1] || r.origin.a[1]>boxmax.a[1] ) return -1;
        
    }
        
    // calculate z time interval
    if (r.direction.a[2]>0) {

        DBL t = (-r.origin.a[2]+boxmax.a[2]) / r.direction.a[2];  // z max
        if (t<tmin) return -1;  if (t<tmax) tmax = t;
        t = (-r.origin.a[2]+boxmin.a[2]) / r.direction.a[2];  // z min
        if (t>tmax) return -1;  if (t>tmin) tmin = t;
            
    } else if(r.direction.a[2]!=0) {

        DBL t = (-r.origin.a[2]+boxmin.a[2]) / r.direction.a[2]; // z max            
        if (t<tmin) return -1;  if (t<tmax) tmax = t;
        t = (-r.origin.a[2]+boxmax.a[2]) / r.direction.a[2]; // z min
        if (t>tmax) return -1;  if (t>tmin) tmin = t;
            
    } else {
        
        if (r.origin.a[2]>boxmin.a[2] || r.origin.a[2]<boxmax.a[2] ) return -1;
        
    }

    stat.success();
    return tmin;

}


DBL BoundingBox::intersect(const Ray &r, DBL tmax) const {

    stat.eval();

    // returns INTERSECTION_TIME_EPSILON if ray's origin inside box!

    // calculate x time interval
    DBL tmin;

    if (r.direction.a[0]>0) {

        DBL t = (-r.origin.a[0]+boxmax.a[0]) / r.direction.a[0];
        if (t<INTERSECTION_TIME_EPSILON) return -1;  if (t<tmax) tmax = t;
        tmin = (-r.origin.a[0]+boxmin.a[0]) / r.direction.a[0];
        if (tmin<INTERSECTION_TIME_EPSILON) tmin = INTERSECTION_TIME_EPSILON;
        else if (tmin>tmax) return -1;

    } else if (r.direction.a[0]!=0){

        DBL t = (-r.origin.a[0]+boxmin.a[0]) / r.direction.a[0];
        if (t<INTERSECTION_TIME_EPSILON) return -1;  if (t<tmax) tmax = t;
        tmin = (-r.origin.a[0]+boxmax.a[0]) / r.direction.a[0];
        if (tmin<INTERSECTION_TIME_EPSILON) tmin = INTERSECTION_TIME_EPSILON;
        else if (tmin>tmax) return -1;
            
    } else {

        if (r.origin.a[0]<boxmin.a[0] || r.origin.a[0]>boxmax.a[0] ) return -1;
        tmin = INTERSECTION_TIME_EPSILON;
            
    }

    // calculate y time interval
    if (r.direction.a[1]>0) {

        DBL t = (-r.origin.a[1]+boxmax.a[1]) / r.direction.a[1]; //y max
        if (t<tmin) return -1;  if (t<tmax) tmax = t;
        t = (-r.origin.a[1]+boxmin.a[1]) / r.direction.a[1]; //y min
        if (t>tmax) return -1;  if (t>tmin) tmin = t;

    } else if (r.direction.a[1]!=0) {

        DBL t = (-r.origin.a[1]+boxmin.a[1]) / r.direction.a[1]; //y max
        if (t<tmin) return -1;  if (t<tmax) tmax = t;
        t = (-r.origin.a[1]+boxmax.a[1]) / r.direction.a[1]; //y min
        if (t>tmax) return -1;  if (t>tmin) tmin = t;

    } else {
        
        if (r.origin.a[1]<boxmin.a[1] || r.origin.a[1]>boxmax.a[1] ) return -1;
    }

    // calculate z time interval
    if (r.direction.a[2]>0) {

        DBL t = (-r.origin.a[2]+boxmax.a[2]) / r.direction.a[2];  // z max
        if (t<tmin) return -1;  if (t<tmax) tmax = t;
        t = (-r.origin.a[2]+boxmin.a[2]) / r.direction.a[2];  // z min
        if (t>tmax) return -1;  if (t>tmin) tmin = t;

    } else if(r.direction.a[2]!=0) {

        DBL t = (-r.origin.a[2]+boxmin.a[2]) / r.direction.a[2]; // z max
        if (t<tmin) return -1;  if (t<tmax) tmax = t;
        t = (-r.origin.a[2]+boxmax.a[2]) / r.direction.a[2]; // z min
        if (t>tmax) return -1;  if (t>tmin) tmin = t;

    } else {
        
        if (r.origin.a[2]>boxmin.a[2] || r.origin.a[2]<boxmax.a[2] ) return -1;
        
    }
    
    stat.success();
    
    return tmin;

}


bool BoundingBox::isInsideAt(const Vector3& a) const {

    return (a.a[0]>=boxmin.a[0] && a.a[1]>=boxmin.a[1] && a.a[2]>=boxmin.a[2] &&
            a.a[0]<=boxmax.a[0] && a.a[1]<=boxmax.a[1] && a.a[2]<=boxmax.a[2]);

}
